#include "motor.h"
#include <stdio.h>

#define PWM_PERIOD

TIM_HandleTypeDef *p_htim;

typedef struct MOTOR_CHANNEL_ {
	int channel;
	GPIO_TypeDef * port;
	uint16_t pin;
} MOTOR_CHANNEL;

typedef struct MOTOR_PINS_ {
	MOTOR_CHANNEL fore;
	MOTOR_CHANNEL back;
} MOTOR_PINS;

MOTOR_PINS motor_pins[4] = {
		{ {TIM_CHANNEL_1, GPIOB, MOTOR_1_FORE_Pin}, {TIM_CHANNEL_1, GPIOD, MOTOR_1_BACK_Pin} },
		{ {TIM_CHANNEL_2, GPIOB, MOTOR_2_FORE_Pin}, {TIM_CHANNEL_2, GPIOD, MOTOR_2_BACK_Pin} },
		{ {TIM_CHANNEL_3, GPIOB, MOTOR_3_FORE_Pin}, {TIM_CHANNEL_3, GPIOD, MOTOR_3_BACK_Pin} },
		{ {TIM_CHANNEL_4, GPIOB, MOTOR_4_FORE_Pin}, {TIM_CHANNEL_4, GPIOD, MOTOR_4_BACK_Pin} }
};

// モータの初期設定
void motor_init(TIM_HandleTypeDef *p_htim_) {
	p_htim = p_htim_;
	__HAL_RCC_GPIOB_CLK_ENABLE();
	__HAL_RCC_GPIOD_CLK_ENABLE();
	HAL_TIM_PWM_Start(p_htim, TIM_CHANNEL_1);
	HAL_TIM_PWM_Start(p_htim, TIM_CHANNEL_2);
	HAL_TIM_PWM_Start(p_htim, TIM_CHANNEL_3);
	HAL_TIM_PWM_Start(p_htim, TIM_CHANNEL_4);
}

// モータの有効化
void motor_enable() {
	HAL_GPIO_WritePin(GPIOB, GPIO_PIN_2, GPIO_PIN_SET);
}

// モータの無効化
void motor_disable() {
	HAL_GPIO_WritePin(GPIOB, GPIO_PIN_2, GPIO_PIN_RESET);
}

// モータのPWM出力と反対側ブリッジの固定出力の設定。
int motor_set_pwm_gpio(MOTOR_CHANNEL pwm, float pwm_fval, MOTOR_CHANNEL gpio,
		int c_val) {

	if (pwm_fval < 0 || 1 < pwm_fval) {
		return 0;
	}
	if (c_val != 0 && c_val != 1) {
		return 0;
	}

	uint16_t pwm_val = pwm_fval*(p_htim->Init.Period);
	GPIO_PinState gpio_val = (c_val==1) ? GPIO_PIN_SET : GPIO_PIN_RESET;

	HAL_GPIO_WritePin(gpio.port, gpio.pin, gpio_val);

	GPIO_InitTypeDef GPIO_InitStruct = { 0 };

	GPIO_InitStruct.Pin = gpio.pin;
	GPIO_InitStruct.Mode = GPIO_MODE_OUTPUT_PP;
	GPIO_InitStruct.Pull = GPIO_NOPULL;
	GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_LOW;
	HAL_GPIO_Init(gpio.port, &GPIO_InitStruct);

	GPIO_InitStruct.Pin = pwm.pin;
	GPIO_InitStruct.Mode = GPIO_MODE_AF_PP;
	GPIO_InitStruct.Pull = GPIO_NOPULL;
	GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_LOW;
	GPIO_InitStruct.Alternate = GPIO_AF2_TIM4;
	HAL_GPIO_Init(pwm.port, &GPIO_InitStruct);

	__HAL_TIM_SET_COMPARE(p_htim, pwm.channel, pwm_val);

	return 1;
}

// モータmotor_iの前進出力をpwm_fval=0.0〜1.0で設定。反対側ブリッジをc_val=[0,1]で設定。
int motor_fore_pwm(MOTOR_n motor_i, float pwm_fval, int c_val) {

	MOTOR_PINS pins = motor_pins[motor_i];

	return motor_set_pwm_gpio(pins.fore, pwm_fval, pins.back, c_val);
}

// モータmotor_iの反転出力をpwm_fval=0.0〜1.0で設定。反対側ブリッジをc_val=[0,1]で設定。
int motor_back_pwm(MOTOR_n motor_i, float pwm_fval, int c_val) {

	MOTOR_PINS pins = motor_pins[motor_i];

	return motor_set_pwm_gpio(pins.back, pwm_fval, pins.fore, c_val);
}

// モータmotor_iの前進・反転出力をpwm_fore_val,pwm_back=0.0〜1.0で設定。
// ただし、pwm_fore_val,pwm_back_valの少なくとも一方は0.0または1.0でなければならない。
int motor_pwm(MOTOR_n motor_i, float pwm_fore_fval, float pwm_back_fval) {

	if (pwm_back_fval == 0.0f ) {
		return motor_fore_pwm(motor_i, pwm_fore_fval, 0);
	}
	if (pwm_fore_fval == 0.0f ) {
		return motor_back_pwm(motor_i, pwm_back_fval, 0);
	}
	if (pwm_back_fval == 1.0f ) {
		return motor_fore_pwm(motor_i, pwm_fore_fval, 1);
	}
	if (pwm_fore_fval == 1.0f ) {
		return motor_back_pwm(motor_i, pwm_back_fval, 1);
	}
	return 0;
}

// モータの動作確認（4つすべて）
void motor4_test(int speedCount[4], uint8_t ser) {

	printf("DUTY Pin: ENCODER SPEED ENC/SPD DUTY/SPD\r\n");

#if 0
	// 値の記録用変数
	struct _vals {
		int i;     // 番号
		int enc;   // エンコーダ
		int speed; // スピード
		int cv;    // エンコーダ値とスピード値のずれ
		int vdiff; // PWMとスピード値のずれ
	} vals[40];
#endif

	// モータドライバの有効化
	motor_enable();

	// モータ動作ループ（各モータの正転および各モータの反転）
	for(int i=0;i<40;i++) {

		// 処理のスキップ
		while(i%20<3) i++;
		while(4<i%20 && i%20<18) i++;

		// デューティ比の設定
		if(i<20) {
			for(int j=0;j<4;j++) {
				motor_pwm(MOTOR_1+j, 0.05*(i+1), 0);
			}
		} else {
			for(int j=0;j<4;j++) {
				motor_pwm(MOTOR_1+j, 0, 0.05*(i-20+1));
			}
		}

		// 時間待ち
		HAL_Delay(200);

		// エンコーダ、速度カウンタの初期値の取得
		int c0[4], v0[4];
		c0[0] = ENC_LF->CNT;
		c0[1] = ENC_LB->CNT;
		c0[2] = ENC_RF->CNT;
		c0[3] = ENC_RB->CNT;
		for(int j=0;j<4;j++) {
			v0[j] = speedCount[j];
		}

		// モータを210ms間動作させる
		HAL_Delay(210);

		// エンコーダ、速度カウンタのカウント数の取得
		c0[0] = ENC_LF->CNT - c0[0];
		c0[1] = ENC_LB->CNT - c0[1];
		c0[2] = ENC_RF->CNT - c0[2];
		c0[3] = ENC_RB->CNT - c0[3];
		for(int j=0;j<4;j++) {
			v0[j] = speedCount[j] - v0[j];
		}

		// 各モータについて値の確認
		for(int j=0;j<4;j++) {

			if(c0[j]>32767) c0[j] -= 65536;
			if(c0[j]<-32767) c0[j] += 65536;
			if(v0[j]<0) v0[j] += 65536;

			// エンコーダとSPEEDの整合性確認用変数
			int16_t cv = c0[j];
			if(cv<0) cv *= -1;
			cv = cv - v0[j]*2;

			// 回転数とPWMの関係の確認用変数
			int vdiff = v0[j] - ((i%20+1)*600)/20;

			// 値の出力
			printf("%4d  J%d: %7d %5d %7d %8d\r\n", 5*(i%20+1), j+1, c0[j], v0[j], cv, vdiff);

#if 0
			// 値の記録
			vals[i].i = i;
			vals[i].enc = c0;
			vals[i].speed = v0;
			vals[i].cv = cv;
			vals[i].vdiff = vdiff;
#endif

			// エンコーダとSPEEDの整合性確認
			if(i%20>5&&(cv>2||cv<-2)) {
				printf("# エンコーダ/SPEED不整合？: J%d (%d)\r\n", j+1, cv);
				if(!ser) {
					for(int k=0;k<4;k++) {
						motor_pwm(MOTOR_1+k, 0, 0);
					}
					while(1) {
						motor_disable();
						HAL_Delay(100);
						motor_enable();
						HAL_Delay(100);
						motor_disable();
						HAL_Delay(100);
						motor_enable();
						HAL_Delay(500);
					}
				}
			}

			// 回転数とPWMの関係の確認
			if((i==4&&(c0[j]<200||1000<c0[j]))||(i==19&&(c0[j]<1200||1400<c0[j]))||
					(i==24&&(-c0[j]<200||1000<-c0[j]))||(i==39&&(-c0[j]<1200||1400<-c0[j]))) {
				printf("# デューティ比/回転数異常？: J%d (%d)\r\n", j+1, vdiff);
				if(!ser) {
					for(int k=0;k<4;k++) {
						motor_pwm(MOTOR_1+k, 0, 0);
					}
					while(1) {
						motor_disable();
						HAL_Delay(100);
						motor_enable();
						HAL_Delay(100);
					}
				}
			}
		}

		if(i==19||i==39) {

			// ブレーキ
			for(int j=0;j<4;j++) {
				motor_pwm(MOTOR_1+j, 1, 1);
			}

			// 時間待ち
			HAL_Delay(100);
		}
	}

	printf("# ser: %d\r\n", ser);
#if 0
	for(int i=0;i<40;i++) {
		printf("%d,%d,%d,%d,%d\r\n", vals[i].i, vals[i].enc, vals[i].speed, vals[i].cv, vals[i].vdiff);
	}
#endif

}

// テスト動作その0
void motor_test0() {
	int sleep = 500;

	motor_pwm(MOTOR_1, 0, 0);
	motor_pwm(MOTOR_2, 0, 0);
	motor_pwm(MOTOR_3, 0, 0);
	motor_pwm(MOTOR_4, 0, 0);

	motor_fore_pwm(MOTOR_1, 1, 0);
	HAL_Delay(sleep);
	motor_fore_pwm(MOTOR_1, 0, 0);
	HAL_Delay(sleep);
	motor_back_pwm(MOTOR_1, 1, 0);
	HAL_Delay(sleep);
	motor_back_pwm(MOTOR_1, 0, 0);
	HAL_Delay(sleep);

	motor_fore_pwm(MOTOR_2, 1, 0);
	HAL_Delay(sleep);
	motor_fore_pwm(MOTOR_2, 0, 0);
	HAL_Delay(sleep);
	motor_back_pwm(MOTOR_2, 1, 0);
	HAL_Delay(sleep);
	motor_back_pwm(MOTOR_2, 0, 0);
	HAL_Delay(sleep);

	motor_fore_pwm(MOTOR_3, 1, 0);
	HAL_Delay(sleep);
	motor_fore_pwm(MOTOR_3, 0, 0);
	HAL_Delay(sleep);
	motor_back_pwm(MOTOR_3, 1, 0);
	HAL_Delay(sleep);
	motor_back_pwm(MOTOR_3, 0, 0);
	HAL_Delay(sleep);

	motor_fore_pwm(MOTOR_4, 1, 0);
	HAL_Delay(sleep);
	motor_fore_pwm(MOTOR_4, 0, 0);
	HAL_Delay(sleep);
	motor_back_pwm(MOTOR_4, 1, 0);
	HAL_Delay(sleep);
	motor_back_pwm(MOTOR_4, 0, 0);
	HAL_Delay(sleep);
}

// テスト動作その1
void motor_test1() {
	int sleep = 1000;

	// 前進
	motor_pwm(MOTOR_1, 1, 0);
	motor_pwm(MOTOR_2, 1, 0);
	motor_pwm(MOTOR_3, 0, 1);
	motor_pwm(MOTOR_4, 0, 1);
	HAL_Delay(sleep);

	// 停止
	motor_pwm(MOTOR_1, 0, 0);
	motor_pwm(MOTOR_2, 0, 0);
	motor_pwm(MOTOR_3, 0, 0);
	motor_pwm(MOTOR_4, 0, 0);
	HAL_Delay(sleep);

	// 後退
	motor_pwm(MOTOR_1, 0, 1);
	motor_pwm(MOTOR_2, 0, 1);
	motor_pwm(MOTOR_3, 1, 0);
	motor_pwm(MOTOR_4, 1, 0);
	HAL_Delay(sleep);

	// 停止
	motor_pwm(MOTOR_1, 0, 0);
	motor_pwm(MOTOR_2, 0, 0);
	motor_pwm(MOTOR_3, 0, 0);
	motor_pwm(MOTOR_4, 0, 0);
	HAL_Delay(sleep);

	// 右斜め前
	motor_pwm(MOTOR_1, 1, 0);
	motor_pwm(MOTOR_2, 0, 0);
	motor_pwm(MOTOR_3, 0, 0);
	motor_pwm(MOTOR_4, 0, 1);
	HAL_Delay(sleep);

	// 停止
	motor_pwm(MOTOR_1, 0, 0);
	motor_pwm(MOTOR_2, 0, 0);
	motor_pwm(MOTOR_3, 0, 0);
	motor_pwm(MOTOR_4, 0, 0);
	HAL_Delay(sleep);

	// 左斜め後ろ
	motor_pwm(MOTOR_1, 0, 1);
	motor_pwm(MOTOR_2, 0, 0);
	motor_pwm(MOTOR_3, 0, 0);
	motor_pwm(MOTOR_4, 1, 0);
	HAL_Delay(sleep);

	// 停止
	motor_pwm(MOTOR_1, 0, 0);
	motor_pwm(MOTOR_2, 0, 0);
	motor_pwm(MOTOR_3, 0, 0);
	motor_pwm(MOTOR_4, 0, 0);
	HAL_Delay(sleep);

	// 左斜め前
	motor_pwm(MOTOR_1, 0, 0);
	motor_pwm(MOTOR_2, 1, 0);
	motor_pwm(MOTOR_3, 0, 1);
	motor_pwm(MOTOR_4, 0, 0);
	HAL_Delay(sleep);

	// 停止
	motor_pwm(MOTOR_1, 0, 0);
	motor_pwm(MOTOR_2, 0, 0);
	motor_pwm(MOTOR_3, 0, 0);
	motor_pwm(MOTOR_4, 0, 0);
	HAL_Delay(sleep);

	// 右斜め後ろ
	motor_pwm(MOTOR_1, 0, 0);
	motor_pwm(MOTOR_2, 0, 1);
	motor_pwm(MOTOR_3, 1, 0);
	motor_pwm(MOTOR_4, 0, 0);
	HAL_Delay(sleep);

	// 停止
	motor_pwm(MOTOR_1, 0, 0);
	motor_pwm(MOTOR_2, 0, 0);
	motor_pwm(MOTOR_3, 0, 0);
	motor_pwm(MOTOR_4, 0, 0);
	HAL_Delay(sleep);

	// 右
	motor_pwm(MOTOR_1, 1, 0);
	motor_pwm(MOTOR_2, 0, 1);
	motor_pwm(MOTOR_3, 1, 0);
	motor_pwm(MOTOR_4, 0, 1);
	HAL_Delay(sleep);

	// 停止
	motor_pwm(MOTOR_1, 0, 0);
	motor_pwm(MOTOR_2, 0, 0);
	motor_pwm(MOTOR_3, 0, 0);
	motor_pwm(MOTOR_4, 0, 0);
	HAL_Delay(sleep);

	// 左
	motor_pwm(MOTOR_1, 0, 1);
	motor_pwm(MOTOR_2, 1, 0);
	motor_pwm(MOTOR_3, 0, 1);
	motor_pwm(MOTOR_4, 1, 0);
	HAL_Delay(sleep);

	// 停止
	motor_pwm(MOTOR_1, 0, 0);
	motor_pwm(MOTOR_2, 0, 0);
	motor_pwm(MOTOR_3, 0, 0);
	motor_pwm(MOTOR_4, 0, 0);
	HAL_Delay(sleep);
}

// テスト動作その2
void motor_test2() {
    for(int i=0; i<=10; i++) {
        motor_pwm(MOTOR_1, 0.1*i, 0);
        motor_pwm(MOTOR_2, 0.1*i, 0);
        motor_pwm(MOTOR_3, 0.1*i, 0);
        motor_pwm(MOTOR_4, 0.1*i, 0);
        HAL_Delay(100);
    }
    HAL_Delay(2000);

    motor_pwm(MOTOR_1, 1, 1);
    motor_pwm(MOTOR_2, 1, 1);
    motor_pwm(MOTOR_3, 1, 1);
    motor_pwm(MOTOR_4, 1, 1);
    HAL_Delay(1000);

    for(int i=0; i<=10; i++) {
        motor_pwm(MOTOR_1, 0, 0.1*i);
        motor_pwm(MOTOR_2, 0, 0.1*i);
        motor_pwm(MOTOR_3, 0, 0.1*i);
        motor_pwm(MOTOR_4, 0, 0.1*i);
        HAL_Delay(100);
    }
    HAL_Delay(2000);

    motor_pwm(MOTOR_1, 1, 1);
    motor_pwm(MOTOR_2, 1, 1);
    motor_pwm(MOTOR_3, 1, 1);
    motor_pwm(MOTOR_4, 1, 1);
    HAL_Delay(1000);

    for(int i=0; i<=10; i++) {
        motor_pwm(MOTOR_1, 0.1*i, 0);
        motor_pwm(MOTOR_2, 0.1*i, 0);
        motor_pwm(MOTOR_3, 0.1*i, 0);
        motor_pwm(MOTOR_4, 0.1*i, 0);
        HAL_Delay(100);
    }
    HAL_Delay(2000);

    motor_pwm(MOTOR_1, 0, 0);
    motor_pwm(MOTOR_2, 0, 0);
    motor_pwm(MOTOR_3, 0, 0);
    motor_pwm(MOTOR_4, 0, 0);
    HAL_Delay(2000);

    for(int i=0; i<=10; i++) {
        motor_pwm(MOTOR_1, 0, 0.1*i);
        motor_pwm(MOTOR_2, 0, 0.1*i);
        motor_pwm(MOTOR_3, 0, 0.1*i);
        motor_pwm(MOTOR_4, 0, 0.1*i);
        HAL_Delay(100);
    }
    HAL_Delay(2000);

    motor_pwm(MOTOR_1, 0, 0);
    motor_pwm(MOTOR_2, 0, 0);
    motor_pwm(MOTOR_3, 0, 0);
    motor_pwm(MOTOR_4, 0, 0);
    HAL_Delay(2000);
}

// モータ動作チェック
void motor_check() {
	while(1) {
		//motor_test0();
		motor_test1();
		motor_test2();
	}
}
