#include "main.h"
#include "sample.h"
#include <stdio.h>

// シリアル受信用のバッファサイズ
#define UART2RXBUFFSIZE 64
#define UART3RXBUFFSIZE 64
#define RS485RXBUFFSIZE 64

// シリアル受信用のバッファ
uint8_t uart2RxBuff[UART2RXBUFFSIZE];
uint8_t uart3RxBuff[UART3RXBUFFSIZE];
uint8_t rs485RxBuff[RS485RXBUFFSIZE];

// シリアル用の構造体
typedef struct _uart_s {
	UART_HandleTypeDef *huart;
	int rxBuffSize;
	uint8_t *rxBuff;
	int p_rxBuff;
} uart_s;

// シリアル用変数
uart_s uart2;
uart_s uart3;
uart_s rs485;

// 速度カウンタ用変数
int speedCount[4] = {0, 0, 0, 0};

// ADC値格納変数
uint16_t adcVals[4] = {0, 0, 0, 0};
uint16_t vadcVals[2] = {0, 0};

// CAN用変数
CAN_TxHeaderTypeDef canTxHeader;
CAN_RxHeaderTypeDef canRxHeader;
uint8_t canTxData[8];
uint8_t canRxData[8];
uint32_t canTxMailbox;
uint8_t canRecvFlag = 0;

// ユーザスイッチ用カウンタ
uint8_t swCount = 0;

// 現在のシリアル受信数の確認
int uartRxNum(uart_s *uart) {
#ifdef STM32L4A6xx
	int index = uart->huart->hdmarx->Instance->CNDTR;
#else
	int index = uart->huart->hdmarx->Instance->NDTR;
#endif
	index = uart->rxBuffSize - index;
	int remainData = index - uart->p_rxBuff;
	if(remainData < 0) remainData += uart->rxBuffSize;
	return remainData;
}

// シリアルで受信したデータを1バイト取得
char uartGetc(uart_s *uart) {
	while(1) {
		if(uartRxNum(uart)>0) break;
	}
	char c = uart->rxBuff[uart->p_rxBuff];
	uart->p_rxBuff++;
	if(uart->p_rxBuff==uart->rxBuffSize) uart->p_rxBuff = 0;
	return c;
}

// シリアルで1バイト送信
void uartPutc(uart_s *uart, char c) {
	char msg[1];
	msg[0] = c;
	while(HAL_UART_Transmit(uart->huart, (uint8_t*)msg, 1, 1)!=HAL_OK);
}

// 速度カウンタ用の処理
void HAL_TIM_IC_CaptureCallback(TIM_HandleTypeDef *htim)
{
	for(int i=0;i<4;i++) {
		if((((htim->Channel)>>i)&1)) {
			speedCount[i]++;
		}
	}
}

// CAN受信処理
void HAL_CAN_RxFifo0MsgPendingCallback(CAN_HandleTypeDef *hcan) {
	if(HAL_CAN_GetRxMessage(hcan, CAN_RX_FIFO0, &canRxHeader, canRxData)==HAL_OK) {
		canRecvFlag = 1;
	}
}

// ユーザスイッチによる割り込み処理
void HAL_GPIO_EXTI_Callback(uint16_t GPIO_Pin) {
	if(GPIO_Pin==USER_SW_Pin) {
		swCount++;
	}
}

void sample(UART_HandleTypeDef *huart2, UART_HandleTypeDef *huart3,
		TIM_HandleTypeDef *encLF, TIM_HandleTypeDef *encLB, TIM_HandleTypeDef *encRF, TIM_HandleTypeDef *encRB,
		TIM_HandleTypeDef *speed, TIM_HandleTypeDef *motor_htim,
		ADC_HandleTypeDef *adc, ADC_HandleTypeDef *vadc, DAC_HandleTypeDef *dac,
		I2C_HandleTypeDef *i2c, SPI_HandleTypeDef *spi2, SPI_HandleTypeDef *spi3, SPI_HandleTypeDef *spi14,
		UART_HandleTypeDef *hrs485,  CAN_HandleTypeDef *can) {

	// シリアル関連の初期設定
	uart2.huart = huart2;
	uart2.rxBuffSize = UART2RXBUFFSIZE;
	uart2.rxBuff = uart2RxBuff;
	uart2.p_rxBuff = 0;

	uart3.huart = huart3;
	uart3.rxBuffSize = UART3RXBUFFSIZE;
	uart3.rxBuff = uart3RxBuff;
	uart3.p_rxBuff = 0;

	rs485.huart = hrs485;
	rs485.rxBuffSize = RS485RXBUFFSIZE;
	rs485.rxBuff = rs485RxBuff;
	rs485.p_rxBuff = 0;

	HAL_UART_Receive_DMA(uart2.huart, uart2.rxBuff, uart2.rxBuffSize);
	HAL_UART_Receive_DMA(uart3.huart, uart3.rxBuff, uart3.rxBuffSize);
	HAL_UART_Receive_DMA(rs485.huart, rs485.rxBuff, rs485.rxBuffSize);

	// エンコーダの初期設定
	HAL_TIM_Encoder_Start(encLF, TIM_CHANNEL_ALL);
	HAL_TIM_Encoder_Start(encLB, TIM_CHANNEL_ALL);
	HAL_TIM_Encoder_Start(encRF, TIM_CHANNEL_ALL);
	HAL_TIM_Encoder_Start(encRB, TIM_CHANNEL_ALL);

	// 速度カウンタの初期設定
	HAL_TIM_IC_Start_IT(speed, TIM_CHANNEL_1);
	HAL_TIM_IC_Start_IT(speed, TIM_CHANNEL_2);
	HAL_TIM_IC_Start_IT(speed, TIM_CHANNEL_3);
	HAL_TIM_IC_Start_IT(speed, TIM_CHANNEL_4);

	// モータの初期設定
	motor_init(motor_htim);

	// DACの設定
	HAL_DAC_Start(dac, DAC_CHANNEL_1);
	HAL_DAC_Start(dac, DAC_CHANNEL_2);

	// ADCの設定
#ifdef STM32L4A6xx
	HAL_ADCEx_Calibration_Start(adc, ADC_SINGLE_ENDED);
	HAL_ADCEx_Calibration_Start(vadc, ADC_SINGLE_ENDED);
	if(HAL_ADC_Start_DMA(adc, (uint32_t*)adcVals, 4)!=HAL_OK) Error_Handler();
	if(HAL_ADC_Start_DMA(vadc, (uint32_t*)vadcVals, 2)!=HAL_OK) Error_Handler();
#else
	HAL_ADC_Start_DMA(adc, (uint32_t*)adcVals, 4);
	HAL_ADC_Start_DMA(vadc, (uint32_t*)vadcVals, 2);
#endif

	// RS485の設定
#ifdef STM32L4A6xx
#else
	HAL_GPIO_WritePin(RS485_DE_GPIO_Port, RS485_DE_Pin, GPIO_PIN_SET);
#endif
	HAL_GPIO_WritePin(RS485_RE_GPIO_Port, RS485_RE_Pin, GPIO_PIN_SET);

	// CANの設定
	CAN_FilterTypeDef canFilter;
	canFilter.FilterIdHigh = 0;
	canFilter.FilterIdLow = 0;
	canFilter.FilterMaskIdHigh = 0;
	canFilter.FilterMaskIdLow = 0;
	canFilter.FilterScale = CAN_FILTERSCALE_32BIT;
	canFilter.FilterFIFOAssignment = CAN_FILTER_FIFO0;
	canFilter.FilterBank = 0;
	canFilter.FilterMode = CAN_FILTERMODE_IDMASK;
	canFilter.SlaveStartFilterBank = 14;
	canFilter.FilterActivation = ENABLE;
	if(HAL_CAN_ConfigFilter(can, &canFilter)!=HAL_OK) {
		Error_Handler();
	}

	HAL_CAN_Start(can);
	if(HAL_CAN_ActivateNotification(can, CAN_IT_RX_FIFO0_MSG_PENDING)!=HAL_OK) {
		Error_Handler();
	}

	// デバッガからのシリアル出力の初期化
	setbuf(stdout, NULL);
	printf("Start sample program...\r\n");

	// 動作チェックモード
	// - LED6=MOTOR_ENが点灯して1秒以内にメインボード上のSW3=USER_SWが押された回数で判定。
	//     - 押された回数が1回の時、モータ動作確認モードへ。
	//     - 押された回数が2回の時、9軸センサ(AE-BMX055)動作確認モードへ。
	//     - 押された回数が0回、または3回以上の時、次へ。
	motor_enable();
	HAL_Delay(1000);
	motor_disable();

	// モータ動作確認モード
	if(swCount==1) {

		// シリアル出力
		uint8_t ser = 1;
		// モータの動作チェック
		motor4_test(speedCount, ser);

		// モータドライバの無効化
		motor_disable();
		// 時間待ち
		HAL_Delay(1000);

		// モータドライバの有効化
		motor_enable();
		// モータの動作確認（無限ループ）
		motor_check();
	}

	// 9軸センサ(AE-BMX055)動作確認モード
	else if(swCount==2) {

		// 動作確認モードへの移行確認用にLED6=MOTOR_ENを点灯
		motor_enable();

		// センサのテスト
		bmx055_i2c_test(i2c);
	}

	// メインループ
	while(1) {

		// UART3に受信データがある場合
		if(uartRxNum(&uart3)>0) {

			char c = uartGetc(&uart3);

			if('0'<=c&&c<='9') {
				uartPutc(&uart3, c-'0'+'a');
			}
		}

		// UART2に受信データがある場合
		if(uartRxNum(&uart2)>0) {

			char c = uartGetc(&uart2);

			// UART2送受信確認
			if('0'<=c&&c<='9') {
				uartPutc(&uart2, c+17);
			}

			// モータ動作確認（個別）
			// モータの正転および逆転を各モータについて行い、以下の値を整数で順に返信する。
			// ・現在のループ（0〜8、1バイト）。
			// ・エンコーダでのカウント値（-32768〜32767、2バイト）。
			// ・速度カウンタでのカウント値（0〜65535、2バイト）。
			// 最後に'T'を返信する。
			else if(c=='m') {

				// モータドライバの有効化
				motor_enable();

				// モータ動作ループ（各モータの正転および各モータの反転）
				for(int i=0;i<8;i++) {

					// エンコーダの初期値の取得
					int c0;
					switch(MOTOR_1+i%4) {
					case MOTOR_1:
						c0 = ENC_LF->CNT;
						break;
					case MOTOR_2:
						c0 = ENC_LB->CNT;
						break;
					case MOTOR_3:
						c0 = ENC_RF->CNT;
						break;
					case MOTOR_4:
						c0 = ENC_RB->CNT;
						break;
					}

					// 速度カウンタの初期化
					speedCount[i%4] = 0;

					// モータを210ms間動作させる
					if(i<4) {
						motor_pwm(MOTOR_1+i%4, 0.5, 0);
					} else {
						motor_pwm(MOTOR_1+i%4, 0, 0.5);
					}
					HAL_Delay(210);
					motor_pwm(MOTOR_1+i%4, 0, 0);

					// 速度カウンタでのカウント数の取得
					int v0 = speedCount[i%4];

					// エンコーダでのカウント数の取得
					switch(MOTOR_1+i%4) {
					case MOTOR_1:
						c0 = ENC_LF->CNT - c0;
						break;
					case MOTOR_2:
						c0 = ENC_LB->CNT - c0;
						break;
					case MOTOR_3:
						c0 = ENC_RF->CNT - c0;
						break;
					case MOTOR_4:
						c0 = ENC_RB->CNT - c0;
						break;
					}
					if(c0>32767) c0 -= 65536;
					if(c0<-32767) c0 += 65536;

					// 現在のループ番号の返信
					uartPutc(&uart2, i);
					// エンコーダでのカウント数の返信
					uartPutc(&uart2, c0&0xFF);
					uartPutc(&uart2, c0>>8);
					// 速度カウンタでのカウント数の返信
					uartPutc(&uart2, v0&0xFF);
					uartPutc(&uart2, v0>>8);

					// 時間待ち
					HAL_Delay(100);
				}

				// 'T'の返信
				uartPutc(&uart2, 'T');
			}

			// モータ動作確認（PWM）
			// 全モータの正転および逆転をデューティ比を変化させながら行い、以下の値を整数で順に返信する。
			// ・現在のループ（0〜39、1バイト、カウント開始のタイミングで送信）。
			// ・エンコーダでのカウント値（-32768〜32767、4x2バイト=8バイト）。
			// ・速度カウンタでのカウント値（0〜65535、4x2バイト=8バイト）。
			// 最後に'T'を返信する。
			else if(c=='M') {

				// モータドライバの有効化
				motor_enable();

				// モータ動作ループ（全モータの正転および反転）
				for(int i=0;i<40;i++) {

					// デューティ比の設定
					for(int j=0;j<4;j++) {
						if(i<20) {
							motor_pwm(MOTOR_1+j%4, 0.05*(i+1), 0);
						} else {
							motor_pwm(MOTOR_1+j%4, 0, 0.05*(i-20+1));
						}
					}

					// 時間待ち
					HAL_Delay(200);

					// 現在のループ番号の返信（カウント開始）
					uartPutc(&uart2, i);

					// エンコーダ、速度カウンタの初期値の取得
					int c0[4];
					int v0[4];
					c0[0] = ENC_LF->CNT;
					c0[1] = ENC_LB->CNT;
					c0[2] = ENC_RF->CNT;
					c0[3] = ENC_RB->CNT;
					for(int j=0;j<4;j++) {
						v0[j] = speedCount[j];
					}

					// モータを210ms間動作させる
					HAL_Delay(210);

					// エンコーダ、速度カウンタのカウント数の取得
					c0[0] = ENC_LF->CNT - c0[0];
					c0[1] = ENC_LB->CNT - c0[1];
					c0[2] = ENC_RF->CNT - c0[2];
					c0[3] = ENC_RB->CNT - c0[3];
					for(int j=0;j<4;j++) {
						v0[j] = speedCount[j] - v0[j];
					}

					// エンコーダでのカウント数の返信
					for(int j=0;j<4;j++) {
						if(c0[j]>32767) c0[j] -= 65536;
						if(c0[j]<-32767) c0[j] += 65536;
						uartPutc(&uart2, c0[j]&0xFF);
						uartPutc(&uart2, c0[j]>>8);
					}
					// 速度カウンタでのカウント数の返信
					for(int j=0;j<4;j++) {
						if(v0[j]<0) v0[j] += 65536;
						uartPutc(&uart2, v0[j]&0xFF);
						uartPutc(&uart2, v0[j]>>8);
					}

					if(i==19||i==39) {
						// ブレーキ
						for(int j=0;j<4;j++) {
							motor_pwm(MOTOR_1+j%4, 1, 1);
						}
						// 時間待ち
						HAL_Delay(100);
					}
				}

				// 'T'の返信
				uartPutc(&uart2, 'T');
			}

			// DACの設定
			// 'd'に続いて、2つのチャンネルのDACを1バイトずつ（値は0〜255）設定。
			else if(c=='d') {
				int v1 = uartGetc(&uart2);
				int v2 = uartGetc(&uart2);
				HAL_DAC_SetValue(dac, DAC_CHANNEL_1, DAC_ALIGN_12B_R, v1*16);
				HAL_DAC_SetValue(dac, DAC_CHANNEL_2, DAC_ALIGN_12B_R, v2*16);
			}

			// ADC値の取得
			// ４つのADC入力値を各1バイトで返信。
			else if(c=='a') {

				// ADC値を取得して返信
				for(int i=0;i<4;i++) {
					uint8_t v = adcVals[i]>>4;
					uartPutc(&uart2, v);
				}
			}

			// バッテリー関連ADC値の取得
			// vehicleとcomputerのバッテリー電圧を各1バイトで返信。
			else if(c=='A') {

				// ADCの開始
				HAL_ADC_Start_DMA(vadc, (uint32_t*)vadcVals, 2);

				// ADC値の取得
				uint8_t vehicle = vadcVals[0]>>4;
				uint8_t computer = vadcVals[1]>>4;

				// ADC値の返信
				uartPutc(&uart2, vehicle);
				uartPutc(&uart2, computer);
			}

			// I2C通信
			// 'c'に続いて、以下の値を設定する。
			// ・アドレス（1ビット左シフトしたもの）
			// ・受信データの長さ（バイト）
			// ・送信データの長さ（バイト）
			// ・送信データ
			// 以上を受信してI2C通信をした後、I2Cで受信したデータを返す。
			else if(c=='c') {

				// 送受信バッファ（最大長さ64）
				uint8_t buf[64];

				// アドレスの取得
				int addr = uartGetc(&uart2);
				// 受信データの長さの取得
				int recvLen = uartGetc(&uart2);
				// 送信データの長さの取得
				int sendLen = uartGetc(&uart2);
				// データ長さが64バイト以上の場合は64バイトに切り詰め
				if(sendLen>64) sendLen = 64;
				if(recvLen>64) recvLen = 64;
				// 送信データの取得
				for(int i=0;i<sendLen;i++) {
					buf[i] = uartGetc(&uart2);
				}

				// I2C送信
				HAL_I2C_Master_Transmit(i2c, addr, buf, sendLen, 3000);
				HAL_Delay(70);

				// I2C受信
				HAL_I2C_Master_Receive(i2c, addr|1, buf, recvLen, 3000);

				// I2Cで受信したデータを返す
				for(int i=0;i<recvLen;i++) {
					uartPutc(&uart2, buf[i]);
				}
			}

			// SPI通信
			// 'p'に続いて、以下の値を設定する。
			// ・SPIチャンネル（'1'、'2'、'3'、'4'のいずれか）
			// ・送受信データの長さ（バイト）
			// ・送信データ
			// 以上を受信してSPI通信をした後、SPIで受信したデータを返す。
			else if(c=='p') {

				// 送信バッファ（最大長さ64）
				uint8_t sendBuf[64];
				// 受信バッファ（最大長さ64）
				uint8_t recvBuf[64];

				// SPIチャンネルの取得
				char s = uartGetc(&uart2);
				// 送受信データの長さの取得
				int len = uartGetc(&uart2);
				// データ長さが64バイト以上の場合は64バイトに切り詰め
				if(len>64) len = 64;
				// 送信データの取得
				for(int i=0;i<len;i++) {
					sendBuf[i] = uartGetc(&uart2);
				}

				// SPI通信
				if(s=='2'||s=='1'||s=='4') {
					// スレーブの場合の通信処理
					if(s=='2') {
						while(HAL_OK!=HAL_SPI_TransmitReceive(spi2, sendBuf, recvBuf, len, 1000));
					} else if(s=='1'||s=='4') {
						while(HAL_OK!=HAL_SPI_TransmitReceive(spi14, sendBuf, recvBuf, len, 1000));
					}
				} else if(s=='3') {
					// マスターの場合の通信
					HAL_Delay(10);
					HAL_SPI_TransmitReceive(spi3, sendBuf, recvBuf, len, 1000);
				}

				// SPIで受信したデータを返す
				for(int i=0;i<len;i++) {
					uartPutc(&uart2, recvBuf[i]);
				}
			}

			// RS485通信
			// 'e'に続いて、以下の値を設定する。
			// ・受信データの長さ（バイト）
			// ・送信データの長さ（バイト）
			// ・送信データ
			// 以上を受信してRS485通信をした後、RS485で受信したデータを返す。
			else if(c=='e') {

				// 送受信バッファ（最大長さ64）
				uint8_t buf[64];

				// 受信データの長さの取得
				int recvLen = uartGetc(&uart2);
				// 送信データの長さの取得
				int sendLen = uartGetc(&uart2);
				// データ長さが64バイト以上の場合は64バイトに切り詰め
				if(sendLen>64) sendLen = 64;
				if(recvLen>64) recvLen = 64;
				// 送信データの取得
				for(int i=0;i<sendLen;i++) {
					buf[i] = uartGetc(&uart2);
				}

				// RS485送信
				for(int i=0;i<sendLen;i++) {
					uartPutc(&rs485, buf[i]);
				}
				HAL_Delay(1);

				// RS485受信
				HAL_GPIO_WritePin(RS485_RE_GPIO_Port, RS485_RE_Pin, GPIO_PIN_RESET);
#ifdef STM32L4A6xx
#else
				HAL_GPIO_WritePin(RS485_DE_GPIO_Port, RS485_DE_Pin, GPIO_PIN_RESET);
#endif
				for(int i=0;i<recvLen;i++) {
					buf[i] = uartGetc(&rs485);
				}
				HAL_GPIO_WritePin(RS485_RE_GPIO_Port, RS485_RE_Pin, GPIO_PIN_SET);
#ifdef STM32L4A6xx
#else
				HAL_GPIO_WritePin(RS485_DE_GPIO_Port, RS485_DE_Pin, GPIO_PIN_SET);
#endif

				// RS485で受信したデータを返す
				for(int i=0;i<recvLen;i++) {
					uartPutc(&uart2, buf[i]);
				}
			}

			// CAN通信
			// 'f'に続いて、以下の値を設定する。
			// ・受信データの長さ（バイト）
			// ・送信データの長さ（バイト）
			// ・送信データ
			// 以上を受信してCAN通信をした後、CANで受信したデータを返す。
			else if(c=='f') {

				// 送受信バッファ（最大長さ8）
				uint8_t buf[8];

				// 受信データの長さの取得
				int recvLen = uartGetc(&uart2);
				// 送信データの長さの取得
				int sendLen = uartGetc(&uart2);
				// データ長さが8バイト以上の場合は8バイトに切り詰め
				if(sendLen>8) sendLen = 8;
				if(recvLen>8) recvLen = 8;
				// 送信データの取得
				for(int i=0;i<sendLen;i++) {
					buf[i] = uartGetc(&uart2);
				}

				// CAN送信
				canTxHeader.StdId = 0x111;
				canTxHeader.RTR = CAN_RTR_DATA;
				canTxHeader.IDE = CAN_ID_STD;
				canTxHeader.DLC = sendLen;
				canTxHeader.TransmitGlobalTime = DISABLE;
				for(int i=0;i<sendLen;i++) {
					canTxData[i] = buf[i];
				}
				if(HAL_CAN_AddTxMessage(can, &canTxHeader, canTxData, &canTxMailbox)!=HAL_OK) {
					Error_Handler();
				}

				// 送受信バッファのクリア
				for(int i=0;i<8;i++) {
					buf[i] = ' ';
				}

				// CAN受信
				while(1) {
					if(canRecvFlag) {
						canRecvFlag = 0;
						for(int i=0;i<canRxHeader.DLC;i++) {
							buf[i] = canRxData[i];
						}
						break;
					}
				}

				// CANで受信したデータを返す
				for(int i=0;i<recvLen;i++) {
					uartPutc(&uart2, buf[i]);
				}
			}

			// ユーザスイッチのカウント数の取得
			else if(c=='u') {
				// ユーザスイッチのカウント数を返す
				uartPutc(&uart2, swCount);
			}

			// 静的変数のカウント数の取得
			else if(c=='s') {
				static uint8_t s_count = 0;
				uartPutc(&uart2, s_count);
				s_count++;
			}
		}
	}
}
